from __future__ import print_function, absolute_import, unicode_literals
import copy
import json
import logging

from os import uname
import sys
import argparse
from osbs import set_logging
from osbs.api import OSBS
from osbs.conf import Configuration
from osbs.constants import BUILD_JSON_STORE, DEFAULT_CONFIGURATION_FILE, DEFAULT_CONFIGURATION_SECTION

try:
    # py2
    from urllib2 import HTTPError
except ImportError:
    # py3
    from urllib.error import HTTPError


logger = logging.getLogger('osbs')


def graceful_chain_get(d, *args):
    t = copy.deepcopy(d)
    for arg in args:
        try:
            t = t[arg]
        except (AttributeError, KeyError):
            return None
    return t


def cmd_list_builds(args, osbs):
    builds = osbs.list_builds(namespace=args.namespace)
    format_str = "{name:48} {status:16} {image:64}"
    print(format_str.format(**{"name": "BUILD NAME", "status": "STATUS", "image": "IMAGE NAME"}), file=sys.stderr)
    for build in builds['items']:
        image = graceful_chain_get(build, 'parameters', 'output', 'imageTag')
        if args.USER:
            if not image.startswith(args.USER + "/"):
                continue
        b = {
            "name": build['metadata']['name'],
            "status": build['status'],
            "image": image,
        }
        print(format_str.format(**b))


def cmd_get_build(args, osbs):
    build_json = osbs.get_build(args.BUILD_ID[0], namespace=args.namespace).json
    # FIXME: pretty printing json could be a better idea
    metadata = build_json.get("metadata", {})
    md = metadata.get("annotations", metadata.get("labels", {}))
    dockerfile = md.get("dockerfile", None)
    packages = md.get("rpm-packages", None)
    logs = md.get("logs", None)
    repositories_json = md.get("repositories", None)
    repositories_str = None
    if repositories_json is not None:
        repositories = json.loads(repositories_json)
        repositories_template = """\
Primary

{primary}

Unique

{unique}"""
        repositories_context = {
            "primary": "\n".join(repositories["primary"]),
            "unique": "\n".join(repositories["unique"]),
        }
        repositories_str = repositories_template.format(**repositories_context)

    template = """\
BUILD ID: {build_id}
STATUS: {status}
IMAGE: {image}
DATE: {date}

DOCKERFILE

{dockerfile}

BUILD LOGS

{logs}

PACKAGES

{packages}

REPOSITORIES

{repositories}"""
    context = {
        "build_id": build_json['metadata']['name'],
        "status": build_json['status'],
        "image": graceful_chain_get(build_json, 'parameters', 'output', 'imageTag'),
        "date": build_json['metadata']['creationTimestamp'],
        "dockerfile": dockerfile,
        "logs": logs,
        "packages": packages,
        "repositories": repositories_str,
    }
    print(template.format(**context))


def cmd_build(args, osbs):
    build = osbs.create_build(
        git_uri=args.git_url,
        git_ref=args.git_commit,
        user=args.user,
        component=args.component,
        target=args.target,
        architecture=args.arch,
        namespace=args.namespace
    )
    build_id = build.build_id
    print("Build submitted (%s), watching logs (feel free to interrupt)" % build_id)
    for line in osbs.get_build_logs(build_id, follow=True):
        print(line)


def cmd_build_logs(args, osbs):
    build_id = args.BUILD_ID[0]
    follow = args.follow

    if follow:
        for line in osbs.get_build_logs(build_id, follow=True, namespace=args.namespace):
            print(line)
    else:
        logs = osbs.get_build_logs(build_id, follow=False, namespace=args.namespace)
        print(logs, end="")


def cmd_watch_build(args, osbs):
    build_json = osbs.wait_for_build_to_finish(args.BUILD_ID[0], namespace=args.namespace)


def cmd_get_token(args, osbs):
    token = osbs.get_token()
    print(token)


def cmd_get_user(args, osbs):
    args_username = args.USERNAME
    if args_username is None:
        user_json = osbs.get_user()
    else:
        args_username = args_username[0]
        user_json = osbs.get_user(args_username)
    name = ""
    full_name = ""
    try:
        name = user_json["metadata"]["name"]
    except KeyError:
        logger.error("\"name\" is not in response")
    try:
        full_name = user_json["fullName"]
    except KeyError:
        logger.error("\"full name\" is not in response")
    print("Name: \"%s\"\nFull Name: \"%s\'" % (name, full_name))


def cli():
    parser = argparse.ArgumentParser(
        description="OpenShift Build Service client"
    )
    exclusive_group = parser.add_mutually_exclusive_group()
    exclusive_group.add_argument("--verbose", action="store_true", default=None)
    exclusive_group.add_argument("-q", "--quiet", action="store_true")

    subparsers = parser.add_subparsers(help='commands')

    list_builds_parser = subparsers.add_parser('list-builds', help='list builds in OSBS',
                                               description="list all builds in specified namespace "
                                               "(to list all builds in all namespaces, use --namespace=\"\")")
    list_builds_parser.add_argument("USER", help="list builds only for specified username",
                                    nargs="?")
    list_builds_parser.set_defaults(func=cmd_list_builds)

    watch_build_parser = subparsers.add_parser('watch-build', help='wait till build finishes')
    watch_build_parser.add_argument("BUILD_ID", help="build ID", nargs=1)
    watch_build_parser.set_defaults(func=cmd_watch_build)

    get_build_parser = subparsers.add_parser('get-build', help='get info about build')
    get_build_parser.add_argument("BUILD_ID", help="build ID", nargs=1)
    get_build_parser.set_defaults(func=cmd_get_build)

    get_token_parser = subparsers.add_parser('get-token', help='get authentication token')
    get_token_parser.set_defaults(func=cmd_get_token)

    get_user_parser = subparsers.add_parser('get-user', help='get info about user')
    get_user_parser.add_argument("USERNAME", nargs="?", default=None)
    get_user_parser.set_defaults(func=cmd_get_user)

    build_logs_parser = subparsers.add_parser('build-logs', help='get or follow build logs')
    build_logs_parser.add_argument("BUILD_ID", help="build ID", nargs=1)
    build_logs_parser.add_argument("-f", "--follow", help="follow logs as they come", action="store_true",
                                   default=False)
    build_logs_parser.set_defaults(func=cmd_build_logs)

    build_parser = subparsers.add_parser('build', help='build an image in OSBS')
    build_parser.add_argument("--build-type", "-T", action="store", metavar="BUILD_TYPE",
                                   help="build type (prod, simple)")
    build_parser.add_argument("--build-json-dir", action="store", metavar="PATH",
                                   help="directory with build jsons")
    build_parser.add_argument("-g", "--git-url", action='store', metavar="URL",
                                   required=True, help="URL to git repo")
    build_parser.add_argument("--git-commit", action='store', default="master",
                                   help="checkout this commit")
    build_parser.add_argument("-t", "--target", action='store',
                                   help="koji target name")
    build_parser.add_argument("-a", "--arch", action='store', default=uname()[4],
                                   help="build architecture")
    build_parser.add_argument("-u", "--user", action='store', required=True,
                                   help="username (will be image prefix)")
    build_parser.add_argument("-c", "--component", action='store', required=True,
                                   help="name of component")
    build_parser.set_defaults(func=cmd_build)

    parser.add_argument("--openshift-uri", action='store', metavar="URL",
                        help="openshift URL to remote API")
    parser.add_argument("--kubelet-uri", action='store', metavar="URL",
                        help="kubelet URL to remote API")
    parser.add_argument("--registry-uri", action='store', metavar="URL",
                        help="registry where images should be pushed")
    parser.add_argument("--config", action='store', metavar="PATH",
                        help="path to configuration file", default=DEFAULT_CONFIGURATION_FILE)
    parser.add_argument("--instance", "-i", action='store', metavar="SECTION_NAME",
                        help="section within config for requested instance", default=DEFAULT_CONFIGURATION_SECTION)
    parser.add_argument("--username", action='store',
                        help="username within OSBS")
    parser.add_argument("--password", action='store',
                        help="password within OSBS")
    parser.add_argument("--use-kerberos", action='store_true', default=None,
                        help="use kerberos for authentication")
    parser.add_argument("--verify-ssl", action='store_true', default=None,
                        help="verify CA on secure connections")
    parser.add_argument("--with-auth", action="store_true", dest="use_auth", default=None,
                        help="get and supply oauth token with every request")
    parser.add_argument("--without-auth", action="store_false", dest="use_auth", default=None,
                        help="don't supply oauth tokens to requests")
    parser.add_argument("--namespace", help="name of namespace to query against "
                                            "(you may require blank namespace with --namespace=\"\")",
                        metavar="NAMESPACE", action="store", default="default")
    args = parser.parse_args()
    return parser, args


def main():
    parser, args = cli()
    os_conf = Configuration(conf_file=args.config, conf_section=args.instance, cli_args=args)
    build_conf = Configuration(conf_file=args.config, conf_section=args.instance, cli_args=args)

    is_verbose = os_conf.get_verbosity()

    if is_verbose:
        set_logging(level=logging.DEBUG)
        logger.debug("Logging level set to debug")
    elif args.quiet:
        set_logging(level=logging.WARNING)
    else:
        set_logging(level=logging.INFO)

    osbs = OSBS(os_conf, build_conf)

    try:
        args.func(args, osbs)
    except AttributeError as ex:
        if hasattr(args, 'func'):
            raise
        else:
            parser.print_help()
    except KeyboardInterrupt:
        print("Quitting on user request.")
        return -1
    except HTTPError as ex:
        logger.error("HTTP error: %d", ex.getcode())
        return -1
    except Exception as ex:
        if is_verbose:
            raise
        else:
            logger.error("Exception caught: %s", repr(ex))
            return -1

if __name__ == '__main__':
    sys.exit(main())
